@@ -19,24 +19,29 @@ class Driver(object):
             Markov Chain Monte Carlo, "OFTI" for Orbits for the Impatient
         num_secondary_bodies (int): number of secondary bodies in the system.
             Should be at least 1.
-        system_mass (float): mean total mass of the system [M_sol]
+        stellar_or_system_mass (float): mass of the primary star (if fitting for
+            dynamical masses of both components) or total system mass (if
+            fitting using relative astrometry only) [M_sol]
         plx (float): mean parallax of the system [mas]
-        mass_err (float, optional): uncertainty on ``system_mass`` [M_sol]
+        mass_err (float, optional): uncertainty on ``stellar_or_system_mass`` [M_sol]
         plx_err (float, optional): uncertainty on ``plx`` [mas]
         lnlike (str, optional): name of function in ``orbitize.lnlike`` that will
             be used to compute likelihood. (default="chi2_lnlike")
-        system_kwargs (dict, optional): ``restrict_angle_ranges``, ``ref_tau_epoch``,
-            ``results`` for ``orbitize.system.System``.
+        system_kwargs (dict, optional): ``restrict_angle_ranges``, ``tau_ref_epoch``,
+            ``fit_secondary_mass``, ``hipparcos_IAD``, ``gaia``, 
+            ``use_rebound``, ``fitting_basis`` for ``orbitize.system.System``.
         mcmc_kwargs (dict, optional): ``num_temps``, ``num_walkers``, and ``num_threads``
             kwargs for ``orbitize.sampler.MCMC``
 
     Written: Sarah Blunt, 2018
     """
 
-    def __init__(self, input_data, sampler_str,
-                 num_secondary_bodies, system_mass, plx,
-                 mass_err=0, plx_err=0, lnlike='chi2_lnlike',
-                 system_kwargs=None, mcmc_kwargs=None):
+    def __init__(
+        self, input_data, sampler_str,
+        num_secondary_bodies, stellar_or_system_mass, plx,
+        mass_err=0, plx_err=0, lnlike='chi2_lnlike',
+        system_kwargs=None, mcmc_kwargs=None
+    ):
 
         # Read in data
         # Try to interpret input as a filename first
@@ -53,11 +58,15 @@ class Driver(object):
         if system_kwargs is None:
             system_kwargs = {}
 
+        #Check if RV data is included, make sure fit_secondary_mass=True
+        if 'rv' in data_table['quant_type'] and ('fit_secondary_mass' not in system_kwargs or system_kwargs['fit_secondary_mass'] == False):
+            raise Exception('If including RV data in orbit fit, set fit_secondary_mass=True')
+
         if sampler_str == 'OFTI' and ('fit_secondary_mass' in system_kwargs and True == system_kwargs['fit_secondary_mass']):
             raise Exception('Run Astrometry+RV in MCMC for now.')
         # Initialize System object which stores data & sets priors
         self.system = orbitize.system.System(
-            num_secondary_bodies, data_table, system_mass,
+            num_secondary_bodies, data_table, stellar_or_system_mass,
             plx, mass_err=mass_err, plx_err=plx_err, **system_kwargs
         )
 
