@@ -1,21 +1,21 @@
 import numpy as np
 import astropy.units as u
 import astropy.constants as consts
-import sys
 import abc
-import math
 import time
+from astropy.time import Time
 
 import emcee
 import ptemcee
 import multiprocessing as mp
+from multiprocessing import Pool
 
 import orbitize.lnlike
 import orbitize.priors
 import orbitize.kepler
-from orbitize.system import radec2seppa
+from orbitize import cuda_ext
+
 import orbitize.results
-import copy
 import matplotlib.pyplot as plt
 
 class Sampler(abc.ABC):
@@ -57,11 +57,10 @@ class Sampler(abc.ABC):
                 documented in System() above. If M=1, this can be a 1d array.
 
         Returns:
-            lnlikes (float): sum of all log likelihoods of the data given input model
+            float: sum of all log likelihoods of the data given input model
 
         """
         # compute the model based on system params
-        # jitter output from compute model
         model, jitter = self.system.compute_model(params)
 
         # fold data/errors to match model output shape. In particualr, quant1/quant2 are interleaved
@@ -88,6 +87,43 @@ class Sampler(abc.ABC):
 
         if self.custom_lnlike is not None:
             lnlikes_sum += self.custom_lnlike(params)
+        
+        if self.system.hipparcos_IAD is not None:
+
+            # compute Ra/Dec predictions at the Hipparcos IAD epochs
+            raoff_model, deoff_model, _ = self.system.compute_all_orbits(
+                params, epochs=self.system.hipparcos_IAD.epochs_mjd
+            ) 
+
+            raoff_model_hip_epoch, deoff_model_hip_epoch, _ = self.system.compute_all_orbits(
+                params, epochs=Time([1991.25], format='decimalyear').mjd
+            ) 
+
+            # subtract off position of star at reference Hipparcos epoch
+            raoff_model[:,0,:] -= raoff_model_hip_epoch[:,0,:]
+            deoff_model[:,0,:] -= deoff_model_hip_epoch[:,0,:]
+
+            # select body 0 raoff/deoff predictions & feed into Hip IAD lnlike fn
+            lnlikes_sum += self.system.hipparcos_IAD.compute_lnlike(
+                raoff_model[:,0,:], deoff_model[:,0,:], params, self.system.param_idx
+            )
+
+            if self.system.gaia is not None:
+
+                gaiahip_epochs = Time(
+                    [self.system.gaia.hipparcos_epoch, self.system.gaia.gaia_epoch], 
+                    format='decimalyear'
+                ).mjd
+
+                # compute Ra/Dec predictions at the Gaia epoch
+                raoff_model, deoff_model, _ = self.system.compute_all_orbits(
+                    params, epochs=gaiahip_epochs
+                ) 
+
+                # select body 0 raoff/deoff predictions & feed into Gaia module lnlike fn
+                lnlikes_sum += self.system.gaia.compute_lnlike(
+                    raoff_model[:,0,:], deoff_model[:,0,:], params, self.system.param_idx
+                )
 
         return lnlikes_sum
 
@@ -97,21 +133,36 @@ class OFTI(Sampler,):
     OFTI Sampler
 
     Args:
-        like (string): name of likelihood function in ``lnlike.py``
         system (system.System): ``system.System`` object
-        custom_lnlike (func): ability to include an addition custom likelihood function in the fit.
-            the function looks like ``clnlikes = custon_lnlike(params)`` where ``params is a RxM array
-            of fitting parameters, where R is the number of orbital paramters (can be passed in system.compute_model()),
-            and M is the number of orbits we need model predictions for. It returns ``clnlikes`` which is an array of
+        like (string): name of likelihood function in ``lnlike.py``
+        custom_lnlike (func): ability to include an addition custom likelihood 
+            function in the fit. The function looks like 
+            ``clnlikes = custon_lnlike(params)`` where ``params`` is a RxM array
+            of fitting parameters, where R is the number of orbital paramters 
+            (can be passed in system.compute_model()),
+            and M is the number of orbits we need model predictions for. 
+            It returns ``clnlikes`` which is an array of
             length M, or it can be a single float if M = 1.
 
     Written: Isabel Angelo, Sarah Blunt, Logan Pearce, 2018
     """
-
     def __init__(self, system, like='chi2_lnlike', custom_lnlike=None):
 
         super(OFTI, self).__init__(system, like=like, custom_lnlike=custom_lnlike)
-        # pdb.set_trace()
+
+        if (
+            (self.system.hipparcos_IAD is not None) or 
+            (len(self.system.rv[0] > 0))
+        ):
+            raise NotImplementedError(
+                """
+                You can only use OFTI with relative astrometry measurements 
+                (no Hipparcos IAD or RVs... yet). Use MCMC, you overachiever, and
+                settle in for a nice long orbit-fit. (But seriously, if you want 
+                this functionality, let us know!)
+                """
+            )
+
         # compute priors and columns containing ra/dec and sep/pa
         self.priors = self.system.sys_priors
 
@@ -175,12 +226,11 @@ class OFTI(Sampler,):
 
         # create an empty results object
         self.results = orbitize.results.Results(
+            self.system,
             sampler_name=self.__class__.__name__,
             post=None,
             lnlike=None,
-            tau_ref_epoch=self.system.tau_ref_epoch,
-            data=self.system.data_table,
-            num_secondary_bodies=self.system.num_secondary_bodies
+            version_number=orbitize.__version__
         )
 
     def prepare_samples(self, num_samples):
@@ -197,8 +247,6 @@ class OFTI(Sampler,):
             num_samples. This should be passed into ``OFTI.reject()``
         """
 
-        # TODO: modify to work for multi-planet systems
-
         # generate sample orbits
         samples = np.empty([len(self.priors), num_samples])
         for i in range(len(self.priors)):
@@ -207,30 +255,27 @@ class OFTI(Sampler,):
             else: # param is fixed & has no prior
                 samples[i, :] = self.priors[i] * np.ones(num_samples)
 
-        for body_num in np.arange(self.system.num_secondary_bodies):
-            # sma, ecc, inc, argp, lan, tau, plx, mtot = [s for s in samples]
-            ref_ind = 6 * body_num
-            sma = samples[ref_ind,:]
-            ecc = samples[ref_ind + 1,:]
-            inc = samples[ref_ind + 2,:]
-            argp = samples[ref_ind + 3,:]
-            lan = samples[ref_ind + 4,:]
-            tau = samples[ref_ind + 5,:]
-            plx = samples[6 * self.system.num_secondary_bodies,:]
+        # Make Converison to Standard Basis:
+        samples = self.system.basis.to_standard_basis(samples)
+        
+        for body_num in np.arange(self.system.num_secondary_bodies) + 1:
+
+            sma = samples[self.system.basis.standard_basis_idx['sma{}'.format(body_num)],:]
+            ecc = samples[self.system.basis.standard_basis_idx['ecc{}'.format(body_num)],:]
+            inc = samples[self.system.basis.standard_basis_idx['inc{}'.format(body_num)],:]
+            argp = samples[self.system.basis.standard_basis_idx['aop{}'.format(body_num)],:]
+            lan = samples[self.system.basis.standard_basis_idx['pan{}'.format(body_num)],:]
+            tau = samples[self.system.basis.standard_basis_idx['tau{}'.format(body_num)],:]
+            plx = samples[self.system.basis.standard_basis_idx['plx'],:]
             if self.system.fit_secondary_mass:
-                m0 = samples[-1,:]
-                m1 = samples[-1-self.system.num_secondary_bodies+body_num,:]
+                m0 = samples[self.system.basis.standard_basis_idx['m0'],:]
+                m1 = samples[self.system.basis.standard_basis_idx['m{}'.format(body_num)],:]
                 mtot = m0 + m1
             else:
-                mtot = samples[-1,:]
+                mtot = samples[self.system.basis.standard_basis_idx['mtot'],:]
                 m1 = None
             
-            if "gamma" in self.system.labels:
-                gamma = samples[6 * self.system.num_secondary_bodies + 1, :]  # Rob: added gamma and sigma parameters
-            if "sigma" in self.system.labels:
-                sigma = samples[6 * self.system.num_secondary_bodies + 2, :]
-
-            min_epoch = self.epoch_idx[body_num]
+            min_epoch = self.epoch_idx[body_num - 1]
             if min_epoch is None:
                 # Don't need to rotate and scale if no astrometric measurments for this body. Brute force rejection sampling
                 continue
@@ -242,9 +287,9 @@ class OFTI(Sampler,):
             meananno = self.epochs[min_epoch]/period_prescale - tau
 
             # compute sep/PA of generated orbits
-            ra, dec, vc = orbitize.kepler.calc_orbit(
+            ra, dec, _ = orbitize.kepler.calc_orbit(
                 self.epochs[min_epoch], sma, ecc, inc, argp, lan, tau, plx, mtot, 
-                tau_ref_epoch=0, mass_for_Kamp=m1, tau_warning=False
+                tau_ref_epoch=0, mass_for_Kamp=m1
             )
             sep, pa = orbitize.system.radec2seppa(ra, dec) # sep[mas], PA[deg]
 
@@ -278,10 +323,10 @@ class OFTI(Sampler,):
             tau = (self.epochs[min_epoch]/period_new - meananno) % 1
 
             # updates samples with new values of sma, pan, tau
-            samples[ref_ind,:] = sma
-            samples[ref_ind + 3,:] = argp
-            samples[ref_ind + 4,:] = lan
-            samples[ref_ind + 5,:] = tau
+            samples[self.system.basis.standard_basis_idx['sma{}'.format(body_num)],:] = sma
+            samples[self.system.basis.standard_basis_idx['aop{}'.format(body_num)],:] = argp
+            samples[self.system.basis.standard_basis_idx['pan{}'.format(body_num)],:] = lan
+            samples[self.system.basis.standard_basis_idx['tau{}'.format(body_num)],:] = tau
 
         return samples
 
@@ -303,17 +348,42 @@ class OFTI(Sampler,):
                 np.array: the log likelihood values of the accepted orbits.
 
         """
+
         lnp = self._logl(samples)
 
         # we just want the chi2 term for rejection, so compute the Gaussian normalization term and remove it
         errs = np.array([self.system.data_table['quant1_err'],
                          self.system.data_table['quant2_err']]).T
+
         if self.has_corr:
             corrs = self.system.data_table['quant12_corr']
         else:
             corrs = None
         lnp_scaled = lnp - orbitize.lnlike.chi2_norm_term(errs, corrs)
 
+        # account for user-set priors on PAN that were destroyed by scale-and-rotate
+        for body_num in np.arange(self.system.num_secondary_bodies) + 1:
+
+            pan_idx = self.system.basis.standard_basis_idx['pan{}'.format(body_num)]
+
+            pan_prior = self.system.sys_priors[pan_idx]
+            if pan_prior is not orbitize.priors.UniformPrior:
+
+                # apply PAN prior
+                lnp_scaled += pan_prior.compute_lnprob(samples[pan_idx,:])
+
+            # prior is uniform but with different bounds that OFTI expects
+            elif (pan_prior.minval != 0) or (
+                (pan_prior.maxval != np.pi) or (pan_prior.maxval != 2*np.pi)
+            ):
+                
+                samples_outside_pan_prior = np.where(
+                    (samples[pan_idx,:] < pan_prior.minval) or 
+                    (samples[pan_idx,:] > pan_prior.maxval)
+                )[0]
+
+                lnp_scaled[samples_outside_pan_prior] = -np.inf
+
         # reject orbits with probability less than a uniform random number
         random_samples = np.log(np.random.random(len(lnp)))
         saved_orbit_idx = np.where(lnp_scaled > random_samples)[0]
@@ -322,32 +392,27 @@ class OFTI(Sampler,):
 
         return saved_orbits, lnlikes
 
-    def _sampler_process(self, output, total_orbits, num_cores, num_samples=10000, Value=0, lock=None):
+    def _sampler_process(self, output, total_orbits, num_samples=10000, Value=0, lock=None):
         """
         Runs OFTI until it finds the number of total accepted orbits desired.
         Meant to be called by run_sampler.
 
         Args:
             output (manager.Queue): manager.Queue object to store results
-
             total_orbits (int): total number of accepted orbits desired by user
-
-            num_cores(int): the number of cores that _run_sampler_base is being
-                            run in parallel on.
-
             num_samples (int): number of orbits to prepare for OFTI to run
                 rejection sampling on
-
             Value (mp.Value(int)): global counter for the orbits generated
-
             lock: mp.lock object to prevent issues caused by access to shared
                   memory by multiple processes
         Returns:
-            output_orbits (np.array): array of accepted orbits,
-                                      size: total_orbits
+            tuple:
 
-            output_lnlikes (np.array): array of log probabilities,
-                                       size: total_orbits
+                output_orbits (np.array): array of accepted orbits,
+                    size: total_orbits
+
+                output_lnlikes (np.array): array of log probabilities,
+                    size: total_orbits
 
         """
 
@@ -383,27 +448,29 @@ class OFTI(Sampler,):
     def run_sampler(self, total_orbits, num_samples=10000, num_cores=None):
         """
         Runs OFTI in parallel on multiple cores until we get the number of total accepted orbits we want.
+
         Args:
             total_orbits (int): total number of accepted orbits desired by user
             num_samples (int): number of orbits to prepare for OFTI to run
                 rejection sampling on. Defaults to 10000.
             num_cores (int): the number of cores to run OFTI on. Defaults to
-                             number of cores availabe.
+                number of cores availabe.
         Return:
-            output_orbits (np.array): array of accepted orbits. Size: total_orbits.
+            np.array: array of accepted orbits. Size: total_orbits.
 
         Written by: Vighnesh Nagpal(2019)
 
         """
-        if num_cores != 1:
-            if num_cores == None:
-                num_cores = mp.cpu_count()
 
-            results = []
-            # orbits_saved is a global counter for the number of orbits generated
-            orbits_saved = mp.Value('i', 0)
-
-            manager = mp.Manager()
+        if num_cores!=1:
+            if num_cores==None:
+                num_cores=mp.cpu_count()
+            
+            results=[]
+            # orbits_saved is a global counter for the number of orbits generated 
+            orbits_saved=mp.Value('i',0)
+            
+            manager = mp.Manager()            
             output = manager.Queue()
 
             # setup the processes
@@ -413,7 +480,7 @@ class OFTI(Sampler,):
             processes = [
                 mp.Process(
                     target=self._sampler_process,
-                    args=(output, nrun_per_core, num_cores, num_samples,
+                    args=(output, nrun_per_core, num_samples,
                           orbits_saved, lock)
                 ) for x in range(num_cores)
             ]
@@ -449,7 +516,7 @@ class OFTI(Sampler,):
 
             self.results.add_samples(
                 np.array(output_orbits),
-                output_lnlikes, labels=self.system.labels
+                output_lnlikes
             )
             return output_orbits
 
@@ -481,7 +548,7 @@ class OFTI(Sampler,):
 
             self.results.add_samples(
                 np.array(output_orbits),
-                output_lnlikes, labels=self.system.labels
+                output_lnlikes
             )
 
             return output_orbits
@@ -497,23 +564,28 @@ class MCMC(Sampler):
 
     Args:
         system (system.System): system.System object
-        num_temps (int): number of temperatures to run the sampler at. Parallel tempering will be
-            used if num_temps > 1 (default=20)
+        num_temps (int): number of temperatures to run the sampler at. 
+            Parallel tempering will be used if num_temps > 1 (default=20)
         num_walkers (int): number of walkers at each temperature (default=1000)
         num_threads (int): number of threads to use for parallelization (default=1)
         like (str): name of likelihood function in ``lnlike.py``
-        custom_lnlike (func): ability to include an addition custom likelihood function in the fit.
-            the function looks like ``clnlikes = custon_lnlike(params)`` where ``params is a RxM array
-            of fitting parameters, where R is the number of orbital paramters (can be passed in system.compute_model()),
-            and M is the number of orbits we need model predictions for. It returns ``clnlikes`` which is an array of
-            length M, or it can be a single float if M = 1.
-        prev_result_filename (str): if passed a filename to an HDF5 file containing a orbitize.Result data,
-            MCMC will restart from where it left off. 
+        custom_lnlike (func): ability to include an addition custom likelihood 
+            function in the fit. The function looks like 
+            ``clnlikes = custon_lnlike(params)`` where ``params`` is a RxM array 
+            of fitting parameters, where R is the number of orbital paramters 
+            (can be passed in system.compute_model()), and M is the number of 
+            orbits we need model predictions for. It returns ``clnlikes`` 
+            which is an array of length M, or it can be a single float if M = 1.
+        prev_result_filename (str): if passed a filename to an HDF5 file 
+            containing a orbitize.Result data, MCMC will restart from where it 
+            left off. 
 
     Written: Jason Wang, Henry Ngo, 2018
     """
-
-    def __init__(self, system, num_temps=20, num_walkers=1000, num_threads=1, like='chi2_lnlike', custom_lnlike=None, prev_result_filename=None):
+    def __init__(
+        self, system, num_temps=20, num_walkers=1000, num_threads=1, 
+        like='chi2_lnlike', custom_lnlike=None, prev_result_filename=None
+    ):
 
         super(MCMC, self).__init__(system, like=like, custom_lnlike=custom_lnlike)
 
@@ -523,12 +595,11 @@ class MCMC(Sampler):
         
         # create an empty results object
         self.results = orbitize.results.Results(
+            self.system,
             sampler_name=self.__class__.__name__,
             post=None,
             lnlike=None,
-            tau_ref_epoch=system.tau_ref_epoch,
-            data=self.system.data_table,
-            num_secondary_bodies=system.num_secondary_bodies
+            version_number=orbitize.__version__
         )
         
         if self.num_temps > 1:
@@ -540,6 +611,9 @@ class MCMC(Sampler):
         # get priors from the system class. need to remove and record fixed priors
         self.priors = []
         self.fixed_params = []
+
+        self.sampled_param_idx = {}
+        sampled_param_counter = 0
         for i, prior in enumerate(system.sys_priors):
 
             # check for fixed parameters
@@ -547,6 +621,10 @@ class MCMC(Sampler):
                 self.fixed_params.append((i, prior))
             else:
                 self.priors.append(prior)
+                self.sampled_param_idx[self.system.labels[i]] = sampled_param_counter
+                sampled_param_counter += 1
+
+        # initialize walkers initial postions
         self.num_params = len(self.priors)
 
         if prev_result_filename is None:
@@ -592,10 +670,12 @@ class MCMC(Sampler):
         Fills in the missing parameters from the chain that aren't being sampled
 
         Args:
-            sampled_params (np.array): either 1-D array of size = number of sampled params, or 2-D array of shape (num_models, num_params)
+            sampled_params (np.array): either 1-D array of size = number of 
+                sampled params, or 2-D array of shape (num_models, num_params)
 
         Returns:
-            full_params (np.array): same number of dimensions as sampled_params, but with num_params including the fixed parameters
+            np.array: same number of dimensions as sampled_params, 
+                but with num_params including the fixed parameters
         """
         if len(self.fixed_params) == 0:
             # nothing to add
@@ -624,7 +704,6 @@ class MCMC(Sampler):
                 parameters being fit, and M is the number of orbits
                 we need model predictions for. Must be in the same order
                 documented in System() above. If M=1, this can be a 1d array.
-
             include_logp (bool): if True, also include log prior in this function
 
         Returns:
@@ -682,7 +761,63 @@ class MCMC(Sampler):
         # include fixed parameters in posterior
         self.post = self._fill_in_fixed_params(self.post)
 
-    def run_sampler(self, total_orbits, burn_steps=0, thin=1, examine_chains=False, output_filename=None, periodic_save_freq=None):
+    def validate_xyz_positions(self):
+        """
+        If using the XYZ basis, walkers might be initialized in an invalid 
+        region of parameter space. This function fixes that by replacing invalid 
+        positions by new randomly generated positions until all are valid.
+        """
+        if self.system.fitting_basis == 'XYZ':
+            if self.use_pt:
+                all_valid = False
+                while not all_valid:
+                    total_invalids = 0
+                    for temp in range(self.num_temps):
+                        to_stand = self.system.basis.to_standard_basis(self.curr_pos[temp,:,:].T.copy()).T
+
+                        # Get invalids by checking ecc values for each companion
+                        indices = [((i * 6) + 1) for i in range(self.system.num_secondary_bodies)]
+                        invalids = np.where((to_stand[:, indices] < 0.) | (to_stand[:, indices] >= 1.))[0]
+
+                        # Redraw samples for the invalid ones
+                        if len(invalids) > 0:
+                            newpos = []
+                            for prior in self.priors:
+                                randompos = prior.draw_samples(len(invalids))
+                                newpos.append(randompos)
+                            self.curr_pos[temp, invalids, :] = np.stack(newpos).T 
+                            total_invalids += len(invalids)
+                    if total_invalids == 0:
+                        all_valid = True
+                        print('All walker positions validated.')
+            else:
+                all_valid = False
+                while not all_valid:
+                    total_invalids = 0
+                    to_stand = self.system.basis.to_standard_basis(self.curr_pos[:,:].T.copy()).T
+
+                    # Get invalids by checking ecc values for each companion
+                    indices = [((i * 6) + 1) for i in range(self.system.num_secondary_bodies)]
+                    invalids = np.where((to_stand[:, indices] < 0.) | (to_stand[:, indices] >= 1.))[0]                    
+
+                    # Redraw saples for the invalid ones
+                    if len(invalids) > 0:
+                        newpos = []
+                        for prior in self.priors:
+                            randompos = prior.draw_samples(len(invalids))
+                            newpos.append(randompos)
+                        self.curr_pos[invalids, :] = np.stack(newpos).T 
+                        total_invalids += len(invalids)
+                    if total_invalids == 0:
+                        all_valid = True
+                        print('All walker positions validated.')
+
+
+
+    def run_sampler(
+        self, total_orbits, burn_steps=0, thin=1, examine_chains=False, 
+        output_filename=None, periodic_save_freq=None
+    ):
         """
         Runs PT MCMC sampler. Results are stored in ``self.chain`` and ``self.lnlikes``.
         Results also added to ``orbitize.results.Results`` object (``self.results``)
@@ -717,97 +852,94 @@ class MCMC(Sampler):
         if nsteps <= 0:
             raise ValueError("Total_orbits must be greater than num_walkers.")
 
-        if self.use_pt:
-            sampler = ptemcee.Sampler(
-                self.num_walkers, self.num_params, self._logl, orbitize.priors.all_lnpriors,
-                ntemps=self.num_temps, threads=self.num_threads, logpargs=[self.priors, ]
-            )
-        else:
-            if self.num_threads != 1:
-                print('Setting num_threads=1. If you want parallel processing for emcee implemented in orbitize, let us know.')
-                self.num_threads = 1
-
-            sampler = emcee.EnsembleSampler(
-                self.num_walkers, self.num_params, self._logl,
-                kwargs={'include_logp': True}
-            )
-                
-        print("Starting Burn in")
-        for i, state in enumerate(sampler.sample(self.curr_pos, iterations=burn_steps, thin=thin)):
+        with Pool(processes=self.num_threads) as pool: 
             if self.use_pt:
-                self.curr_pos = state[0]
+                sampler = ptemcee.Sampler(
+                    self.num_walkers, self.num_params, self._logl, orbitize.priors.all_lnpriors,
+                    ntemps=self.num_temps, threads=self.num_threads, logpargs=[self.priors, ]
+                )
             else:
-                self.curr_pos = state.coords
+                sampler = emcee.EnsembleSampler(
+                    self.num_walkers, self.num_params, self._logl, pool=pool,
+                    kwargs={'include_logp': True}
+                )
 
-            if (i+1) % 5 == 0:
-                print(str(i+1)+'/'+str(burn_steps)+' steps of burn-in complete', end='\r')
+            print("Starting Burn in")
+            for i, state in enumerate(sampler.sample(self.curr_pos, iterations=burn_steps, thin=thin)):
+                if self.use_pt:
+                    self.curr_pos = state[0]
+                else:
+                    self.curr_pos = state.coords
 
-            if periodic_save_freq is not None:
-                if (i+1) % periodic_save_freq == 0: # we've completed i+1 steps
-                    self.results.curr_pos = self.curr_pos
-                    self.results.save_results(output_filename)
+                if (i+1) % 5 == 0:
+                    print(str(i+1)+'/'+str(burn_steps)+' steps of burn-in complete', end='\r')
 
-        sampler.reset()
-        print('')
-        print('Burn in complete. Sampling posterior now.')
+                if periodic_save_freq is not None:
+                    if (i+1) % periodic_save_freq == 0: # we've completed i+1 steps
+                        self.results.curr_pos = self.curr_pos
+                        self.results.save_results(output_filename)
 
-        saved_upto = 0 # keep track of how many steps of this chain we've saved. this is the next index that needs to be saved 
-        for i, state in enumerate(sampler.sample(self.curr_pos, iterations=nsteps, thin=thin)):
-            if self.use_pt:
-                self.curr_pos = state[0]
-            else:
-                self.curr_pos = state.coords
-                
-            # print progress statement
-            if (i+1) % 5 == 0:
-                print(str(i+1)+'/'+str(nsteps)+' steps completed', end='\r')
+            sampler.reset()
+            print('')
+            print('Burn in complete. Sampling posterior now.')
 
-            if periodic_save_freq is not None:
-                if (i+1) % periodic_save_freq == 0: # we've completed i+1 steps
-                    self._update_chains_from_sampler(sampler, num_steps=i+1)
+            saved_upto = 0 # keep track of how many steps of this chain we've saved. this is the next index that needs to be saved 
+            for i, state in enumerate(sampler.sample(self.curr_pos, iterations=nsteps, thin=thin)):
+                if self.use_pt:
+                    self.curr_pos = state[0]
+                else:
+                    self.curr_pos = state.coords
+                    
+                # print progress statement
+                if (i+1) % 5 == 0:
+                    print(str(i+1)+'/'+str(nsteps)+' steps completed', end='\r')
 
-                    # figure out what is the new chunk of the chain and corresponding lnlikes that have been computed before last save
-                    # grab the current posterior and lnlikes and reshape them to have the Nwalkers x Nsteps dimension again
-                    post_shape = self.post.shape
-                    curr_chain_shape = (self.num_walkers, post_shape[0]//self.num_walkers, post_shape[-1])
-                    curr_chain = self.post.reshape(curr_chain_shape)
-                    curr_lnlike_chain = self.lnlikes.reshape(curr_chain_shape[:2])
-                    # use the reshaped arrays and find the new steps we computed
-                    curr_chunk = curr_chain[:, saved_upto:i+1]
-                    curr_chunk = curr_chunk.reshape(-1, curr_chunk.shape[-1]) # flatten nwalkers x nsteps dim
-                    curr_lnlike_chunk = curr_lnlike_chain[:, saved_upto:i+1].flatten()
+                if periodic_save_freq is not None:
+                    if (i+1) % periodic_save_freq == 0: # we've completed i+1 steps
+                        self._update_chains_from_sampler(sampler, num_steps=i+1)
 
-                    # add this current chunk to the results object (which already has all the previous chunks saved)
-                    self.results.add_samples(curr_chunk, curr_lnlike_chunk, 
-                                                labels=self.system.labels, curr_pos=self.curr_pos)
-                    self.results.save_results(output_filename)
-                    saved_upto = i+1
+                        # figure out what is the new chunk of the chain and corresponding lnlikes that have been computed before last save
+                        # grab the current posterior and lnlikes and reshape them to have the Nwalkers x Nsteps dimension again
+                        post_shape = self.post.shape
+                        curr_chain_shape = (self.num_walkers, post_shape[0]//self.num_walkers, post_shape[-1])
+                        curr_chain = self.post.reshape(curr_chain_shape)
+                        curr_lnlike_chain = self.lnlikes.reshape(curr_chain_shape[:2])
+                        # use the reshaped arrays and find the new steps we computed
+                        curr_chunk = curr_chain[:, saved_upto:i+1]
+                        curr_chunk = curr_chunk.reshape(-1, curr_chunk.shape[-1]) # flatten nwalkers x nsteps dim
+                        curr_lnlike_chunk = curr_lnlike_chain[:, saved_upto:i+1].flatten()
 
-        print('')
-        self._update_chains_from_sampler(sampler)
+                        # add this current chunk to the results object (which already has all the previous chunks saved)
+                        self.results.add_samples(curr_chunk, curr_lnlike_chunk, 
+                                                    curr_pos=self.curr_pos)
+                        self.results.save_results(output_filename)
+                        saved_upto = i+1
 
-        if periodic_save_freq is None:
-            # need to save everything
-            self.results.add_samples(self.post, self.lnlikes, labels=self.system.labels, curr_pos=self.curr_pos)
-        elif saved_upto < nsteps:
-            # just need to save the last few
-            # same code as above except we just need to grab the last few
-            post_shape = self.post.shape
-            curr_chain_shape = (self.num_walkers, post_shape[0]//self.num_walkers, post_shape[-1])
-            curr_chain = self.post.reshape(curr_chain_shape)
-            curr_lnlike_chain = self.lnlikes.reshape(curr_chain_shape[:2])
-            curr_chunk = curr_chain[:, saved_upto:]
-            curr_chunk = curr_chunk.reshape(-1, curr_chunk.shape[-1]) # flatten nwalkers x nsteps dim
-            curr_lnlike_chunk = curr_lnlike_chain[:, saved_upto:].flatten()
+            print('')
+            self._update_chains_from_sampler(sampler)
 
-            self.results.add_samples(curr_chunk, curr_lnlike_chunk, 
-                                                labels=self.system.labels, curr_pos=self.curr_pos)
+            if periodic_save_freq is None:
+                # need to save everything
+                self.results.add_samples(self.post, self.lnlikes, curr_pos=self.curr_pos)
+            elif saved_upto < nsteps:
+                # just need to save the last few
+                # same code as above except we just need to grab the last few
+                post_shape = self.post.shape
+                curr_chain_shape = (self.num_walkers, post_shape[0]//self.num_walkers, post_shape[-1])
+                curr_chain = self.post.reshape(curr_chain_shape)
+                curr_lnlike_chain = self.lnlikes.reshape(curr_chain_shape[:2])
+                curr_chunk = curr_chain[:, saved_upto:]
+                curr_chunk = curr_chunk.reshape(-1, curr_chunk.shape[-1]) # flatten nwalkers x nsteps dim
+                curr_lnlike_chunk = curr_lnlike_chain[:, saved_upto:].flatten()
 
-        if output_filename is not None:
-            self.results.save_results(output_filename)
+                self.results.add_samples(curr_chunk, curr_lnlike_chunk, 
+                                                     curr_pos=self.curr_pos)
 
-        print('Run complete')
+            if output_filename is not None:
+                self.results.save_results(output_filename)
 
+            print('Run complete')
+        # Close pool
         if examine_chains:
             self.examine_chains()
 
@@ -857,10 +989,10 @@ class MCMC(Sampler):
         else:  # build list from user input strings
             params_plot_list = []
             for i in param_list:
-                if i in self.system.param_idx:
-                    params_plot_list.append(self.system.param_idx[i])
+                if i in self.system.basis.param_idx:
+                    params_plot_list.append(self.system.basis.param_idx[i])
                 else:
-                    raise Exception('Invalid param name: {}. See system.param_idx.'.format(i))
+                    raise Exception('Invalid param name: {}. See system.basis.param_idx.'.format(i))
             params_to_plot = np.array(params_plot_list)
 
         # Loop through each parameter and make plot
@@ -879,16 +1011,15 @@ class MCMC(Sampler):
 
     def chop_chains(self, burn, trim=0):
         """
-        Permanently removes steps from beginning (and/or end) of chains from the Results object.
-        Also updates `curr_pos` if steps are removed from the end of the chain
+        Permanently removes steps from beginning (and/or end) of chains from the 
+        Results object. Also updates `curr_pos` if steps are removed from the 
+        end of the chain.
 
         Args:
             burn (int): The number of steps to remove from the beginning of the chains
             trim (int): The number of steps to remove from the end of the chians (optional)
 
-        Returns:
-            None. Updates self.curr_pos and the `Results` object.
-            .. Warning:: Does not update bookkeeping arrays within `MCMC` sampler object.
+        .. Warning:: Does not update bookkeeping arrays within `MCMC` sampler object.
 
         (written): Henry Ngo, 2019
         """
@@ -897,7 +1028,6 @@ class MCMC(Sampler):
         flatchain = np.copy(self.results.post)
         total_samples, n_params = flatchain.shape
         n_steps = np.int(total_samples/self.num_walkers)
-        # TODO: May have to change this to merge with other branches
         flatlnlikes = np.copy(self.results.lnlike)
 
         # Reshape chain to (nwalkers, nsteps, nparams)
@@ -924,13 +1054,53 @@ class MCMC(Sampler):
 
         # Update results object associated with this sampler
         self.results = orbitize.results.Results(
+            self.system, 
             sampler_name=self.__class__.__name__,
             post=flat_chopped_chain,
             lnlike=flat_chopped_lnlikes,
-            tau_ref_epoch=self.system.tau_ref_epoch,
-            labels=self.system.labels,
-            num_secondary_bodies=self.system.num_secondary_bodies
+            version_number = orbitize.__version__,
+            curr_pos = self.curr_pos
         )
 
         # Print a confirmation
         print('Chains successfully chopped. Results object updated.')
+
+    def check_prior_support(self):
+        """
+        Review the positions of all MCMC walkers, to verify that they are supported by the prior space.
+        This function will raise a descriptive ValueError if any positions lie outside prior support.
+        Otherwise, it will return nothing.
+
+        (written): Adam Smith, 2021
+        """
+
+        # Flatten the walker/temperature positions for ease of manipulation.
+        all_positions = self.curr_pos.reshape(self.num_walkers*self.num_temps,self.num_params)
+        
+        # Placeholder list to track any bad parameters that come up.
+        bad_parameters = []
+
+        # If there are no covarient priors, loop on each variable to locate any out-of-place parameters. (this is why we transpose the walkers)
+        if not np.any([prior.is_correlated for prior in self.priors]):
+            for i, x in enumerate(all_positions.T):
+                # Any issues with this parameter?
+                lnprob = self.priors[i].compute_lnprob(np.array(x))
+                supported = np.isfinite(lnprob).all() == True
+
+                if supported == False:
+                    # Problem detected. Take note and continue the loop - we want to catch all the problem parameters.
+                    bad_parameters.append(str(i))
+
+            # Throw our ValueError if necessary,
+            if len(bad_parameters) > 0:
+                raise ValueError("Attempting to start with walkers outside of prior support: check parameter(s) "+', '.join(bad_parameters))
+
+        # We're not done yet, however. There may be errors in covariant priors; run a check for that.
+        else:
+            for y in all_positions:
+                lnprob = orbitize.priors.all_lnpriors(y,self.priors)
+                if not np.isfinite(lnprob).all():
+                    raise ValueError("Attempting to start with walkers outside of prior support: covariant prior failure.")
+        
+        # otherwise exit the function and continue.
+        return
